#include "header.h"

//------------------------ Creation ----------------
Creation::Creation() {
  INDEX sessIDX;
  myIO  *IO;


  // readInit();
  nNodes = (int) IO->readNL();
  nSessions =  (int) IO->readNL();
  maxTrans =   (int) IO->readNL();     // maximum number of transactions
  currTrans =  (int) IO->readNL();

  avgSize =  (int) IO->readNL();
  nQoSLevels =  (int) IO->readNL();
  maxBand =  (int) IO->readNL();
  minBand =  (int) IO->readNL();
  minDelay =  (int) IO->readNL();
  maxDelay =  (int) IO->readNL();
  minHops =  (int) IO->readNL();
  maxHops =  (int) IO->readNL();
  orderedJoinLeave = (BOOL) IO->readNL();

  allSess = new (Session *)[nSessions];

  for(sessIDX = 0; sessIDX < nSessions; sessIDX++)
    {
      allSess[sessIDX] = new Session(sessIDX, nNodes);
      allSess[sessIDX]->setQoS(nQoSLevels, maxBand, minBand, \
			       minDelay, maxDelay, minHops, maxHops); 
    }
   
};



INDEX Creation::selectSess(){
 
  return (INDEX) Probab::uni_inter_Prob(0, nSessions-1,  0);
}

BOOL Creation::doJoin(){
  INDEX sessIDX;
  BOOL ok;

  sessIDX = selectSess();
  ok = getSess(sessIDX)->newJoin();
  return ok;
};

BOOL Creation::doRemove(){
  INDEX sessIDX;
  BOOL ok;

  sessIDX = selectSess();
  ok = getSess(sessIDX)->newRemove();
  return ok;
};

void Creation::printSessions()
{
  INDEX  sessIDX;
  INDEX  src;

	Out <<"// Print Group requests GRP: " << NL;
	Out << "//  node: session: type of request: band:  del:  hops";
	Out << NL;
	Out << "// Average Group Size: " << avgSize;
	Out << NL;
	Out << "// Is Join-Leave ordered? (1=YES ): ";
	Out << getOrderedJoinLeave();
	Out << NL;

    Out << " SESSIONS { " << NL;
    for(sessIDX = 0; sessIDX < nSessions; sessIDX++)
      {
	src = getSess(sessIDX)->getSource();
	Out << "S : " << sessIDX << " : " << src << NL;
      }
    Out << "} " << NL;
    //---------- print initiation of session I: sessIDX ----

    for(sessIDX = 0; sessIDX < nSessions; sessIDX++)
      {
	Out << "GRP_REQ {" << NL;
	Out << "I : " << sessIDX ;
	Out << NL;
	Out << "}" << NL;
      }
    
};

void Creation::run(int argc, char **argv) {
  INDEX  sessIDX;
  INDEX  size;
  BOOL ok;
  BOOL isPrevReqJoin; // Aux: Yes means previous req. was join
  INDEX nJoins, nRemoves; // counters for joins and removes
  INDEX count; // Aux: avoid getting stuck looking for join/remove
  double req_SEED;
  int i;  // aux count the number of removes at the end of the session
  int excess; // number of destinations to be removed

  // ------ printSessions() ------
  printSessions();

    if (argc >= 2)
      req_SEED = atoi(argv[1]);
    else
      req_SEED = 0;
    //Initialize random generator
    Probab::initRandomProb((int) req_SEED);

  
  nJoins = 0;
  nRemoves = 0;
  // ----  BUILD UP of Sessions ----
  Out << "// Start: Beef up session " << NL;
  // Create avgSize-1 joins  since the root is in the group
  for(size = 1; size < avgSize; size++) {
    for(sessIDX = 0; sessIDX < nSessions; sessIDX++) {
      ok = NO;
      ok = getSess(sessIDX)->newJoin();
      if (ok == YES) {
	oneMoreTrans();
        nJoins++;
      }
    } // for
  }   // for

  // --- fluctuations around average ----

  //setOrderedJoinLeave(NO);
  if (getOrderedJoinLeave() == NO)
  {
      isPrevReqJoin = (BOOL) Probab::coinFlip();
  } else 
    {
      isPrevReqJoin = NO;
    }
  Out << "// JOIN-LEAVE  Join-Delete Randomly " << NL;
  while (currTrans < maxTrans) {
    // select session from 0 to nsessions with seed 0
    ok = NO;
    if (isPrevReqJoin == NO)
      {
	count = 0;
	// Repeat join until it succeeds
	//	while (ok == NO) {
	  ok = doJoin();
	  isPrevReqJoin = YES;
	  //count++;
	  // when all sessions have all the nodes

	  //}
	if (ok == YES) {
	  nJoins++;
	  oneMoreTrans();
	  // when no session has participants
	  //	  if (count => nNodes ) {
	  // error( "Count Bailout: Too many attempts "<<NL)
	  // break ;
	  //}
	}
      } else 
	{
	  // repeat doRemove until it succeeds
	  count = 0;
	  //	  while (ok == NO) {
	    ok = doRemove();
	    isPrevReqJoin = NO;
	    // count++;
	    // if I repeat too many times exit
	    //if (count >= nNodes )
	    //break ;
	    //}
	  if (ok == YES) {
	    nRemoves++;
	    oneMoreTrans();
	  }
	}
    if (getOrderedJoinLeave() == NO) {
      isPrevReqJoin = (BOOL) Probab::coinFlip();
    } 
  }

  excess = nJoins-nRemoves;
  Out << " // LEAVE end of session " << NL;
  for(i = 1; i <=  excess; i++) 
    {
        ok = doRemove();
        if (ok == YES) {
          nRemoves++;
          oneMoreTrans();
        } 
    }

  Out << " // Trans: " << currTrans;
  Out << " J: " << nJoins;
  Out << " R: " << nRemoves << NL;
  
}

//------------------------ Session ----------------

 void Session::setQoS( INDEX  levels, CAP minB, CAP maxB, \
		       DELAY minD, DELAY maxD,  HOPS minH, HOPS maxH) {
  nQoSLevels =  levels;
  maxBand =  minB;
  minBand =  maxB;
  minDelay = minD;
  maxDelay = maxD;
  minHops =  minH;
  maxHops =  maxH;

};

Session::Session(INDEX s, INDEX nodes){
  INDEX nIDX;

  id = s;
  nNodes = nodes;

  Parti = new BOOL[nodes];
  for(nIDX = 0; nIDX < nodes; nIDX++)
    setParti(nIDX, NO);

  //  select a source and include it in participants
  source = Probab::uni_inter_Prob(0, nNodes-1, 0) ;
  setParti(source, YES);

};

void Session::print()
{
  Out << " SESS: " << id << " source: " << source << NL;
  
}

BOOL Session::newRemove(){
  BOOL *InSess; // array of nodes  in the session
  INDEX idx;   // index of the node we select in hte InSess[]
  INDEX ni;    // dumb auxiliary
  INDEX count; // count the nodes not in session
  INDEX node;
  BOOL ret_val;  // return value

  count  = 0;
  InSess = new BOOL[nNodes];
  // Identify the nodes in the session except for source
  for(ni=0; ni<nNodes; ni++)
    {
      if ((isInSession(ni) == YES) && (ni != source))
	{
	  count++;
	  InSess[count] = ni;
	}
    }
  // ---  if there are no destinations in the session exit ----
  if (count == 0) {
    ret_val = NO;
  } else
    { //  Select one of the destinations
      //  get an number in [1, count]
      idx = Probab::uni_inter_Prob(1, count, 0);
      node = InSess[idx];
     
      printRemove(node);
      setParti(node, NO);
      ret_val = YES;
    }
  //print();
  delete []InSess;
  return ret_val;
};

void Session::printJoin(INDEX node, CAP band, DELAY del, HOPS hops)
{
  Out << "GRP_REQ {" << NL;
  Out << "I : 0 " << NL  ; // for compatibility for reading the file
  Out << USER_REQ_str;
  Out << " : "  << node;
  Out << " : "  << id ;   // session id
  Out << " : "  << JOIN_CODE ;  ///  JOINCODE is 1
  Out << " : "  << band;
  Out << " : "  << del;
  Out << " : "  << hops;
  Out << NL << "}" << NL;
};

void Session::printRemove(INDEX node)
{
  Out << "GRP_REQ {" << NL;
  Out << "I : 0 " << NL  ; // for compatibility for reading the file
  Out << USER_REQ_str;
  Out << " : "  << node;
  Out << " : "  << id ;   // session id
  Out << " : "  << REMOVE_CODE ;  ///  JOINCODE is 1
  Out << " : "  << 0;   // band
  Out << " : "  << 0;   // delay
  Out << " : "  << 0;   // hops
  Out << NL << "}" << NL;
};


BOOL Session::newJoin(){
  BOOL *OffSess; // array of nodes not in the session
  INDEX idx;   // index of the node we select in hte OffSess[]
  INDEX ni;    // dumb auxiliary
  INDEX count; // count the nodes not in session
  INDEX node;
  CAP band;
  DELAY del;
  HOPS hops;
  BOOL ret_val; //return value

  count  = 0;
  OffSess = new BOOL[nNodes];
  //  for(ni=0; ni<nNodes; ni++)
  //   OffSess[ni] = NO;

  for(ni=0; ni<nNodes; ni++)
    {
      if (isInSession(ni) == NO)
	{
	  OffSess[count] = ni;
	  count++;
	}
    }
  // ---  if all nodes are in the session exit ----
  if (count == 0)
    ret_val = NO;
  else
    {
      //  get an number in [0, count-1]
      idx = Probab::uni_inter_Prob(0, count-1, 0);
      node = OffSess[idx];
  // Choose which Band, Del, Hop 
  // For now uniformly distributed
      band =  Probab::uni_inter_Prob(minBand, maxBand , 0);   // band
      del =  Probab::uni_inter_Prob(minDelay, maxDelay, 0);   // delay
      hops = Probab::uni_inter_Prob(minHops, maxHops, 0);   // hops
      printJoin(node, band, del, hops);
      setParti(node, YES);
      ret_val = YES;
    }

  delete []OffSess;
  return ret_val;
};

// -------------- myIO ---------------

double myIO::readNL()
{ // READ next number and discard the rest of the line
  // ATTENTION: File must be ready to read a number
  char c;
  double num;

  cin >> num;
  do
  {
    cin.get(c);
  }  while( c != '\n' );
  return num;
};
